"""
This library is the core of Youtube Autonomous,
where all the functionalities are joint and the
magic can happen.

This library manages whole projects and is 
capable of creating awesome videos from scratch.
Using AI, video effects, audio narrations and
many more things that you'll discover inside.

- The wizard: danielalcalavalera@gmail.com
"""
from yta_core.database import database_handler
from yta_core.settings import Settings
from yta_core.classes.project_raw import ProjectRaw
from yta_core.classes.project import Project
from yta_core.database.dataclasses import MongoDBProject
from yta_general_utils.programming.decorators import singleton_old
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.logger import print_completed, print_in_progress
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.path import create_file_abspath
from yta_general_utils.temp import Temp


@singleton_old
class YTA:
    """
    The amazing Youtube Autonomous class that is
    able to magically create and edit youtube
    videos by using AI and other functionalities.
    """

    @property
    def unfinished_projects(
        self
    ) -> list[MongoDBProject]:
        """
        Get a list with the unfinished projects from
        the database. This will be empty if there are
        no unfinished projects.
        """
        return [
            MongoDBProject(project.json)
            for project in database_handler.unfinished_projects
        ]

    def __init__(
        self
    ):
        #  TODO: Complete it
        self._segments_abspath = f'{get_project_abspath()}{Settings.DEFAULT_SEGMENT_PARTS_FOLDER}/'
        self._projects_output_abspath = f'{get_project_abspath()}{Settings.DEFAULT_PROJECTS_OUTPUT_FOLDER}/'

        # We force to create the folder if it doesn't exist
        create_file_abspath(f'{self._segments_abspath}toforce')
        create_file_abspath(f'{self._projects_output_abspath}toforce')

    def purge(
        self,
        do_remove_segments_files: bool = False
    ):
        """
        Clean the temporary folder removing all the
        previously generated temporary files and also
        the segment files id the parameter
        'do_remove_segments_files' is set as True.
        """
        ParameterValidator.validate_mandatory_bool('do_remove_segments_files', do_remove_segments_files)

        Temp.clean_folder()

        if do_remove_segments_files:
            # TODO: Remove all files in self._segments_abspath folder
            pass

    def check_config(
        self
    ):
        """
        TODO: What configuration do we mean? That we
        are able to connect to db and that stuff (?)
        """
        # TODO: Check that the config is ok
        pass

    def insert_project_in_database_from_file(
        self,
        filename: str
    ) -> Project:
        """
        Read the provided project content 'filename' and
        create a new project in the database if the provided
        'filename' contains a new project and is valid. If
        the information belongs to an already registered
        project, it will raise an exception.

        This method returns the new stored project mongo
        ObjectId as a string if successfully stored, or
        raises an Exception if anything went wrong.
        """
        ParameterValidator.validate_mandatory_string('filename', filename)

        print('pre init from file')
        project_from_file = ProjectRaw.init_from_file(filename)
        print('post init from file')
        # If we are here, the raw project is valid

        # We don't want to have duplicated projects
        print('pre get raise')
        if database_handler.get_project_by_script(project_from_file) is not None:
            raise Exception('There is an existing project in the database with the same content.')
        print('post get raise')

        # Ok, it doesn't exist, lets transform into a
        # ProjectForDatase and store it
        print('pre insert')
        project = database_handler.insert_project(project_from_file.as_project_validated)
        print('post insert')

        print_completed(f'Project created in database with ObjectId = "{project.id}"')

        return Project(project.id)
    
    def _process_unfinished_projects(
        self
    ) -> bool:
        """
        Get the unfinished projects and process
        them one by one.
        """
        for db_project in self.unfinished_projects:
            print_in_progress(f'Processing project "{db_project.id}"')
            self._process_project(db_project)
            print_completed(f'Project "{db_project.id}" processed succesfully!')

        return True
    
    def _process_project(
        self,
        db_project: MongoDBProject
    ) -> bool:
        """
        Process the provided 'db_project' project, which must
        be a valid project stored in the database.
        """
        filename = f'{self._projects_output_abspath}project_{db_project.id}.mp4'

        return Project(db_project.id).build(filename)


"""
The edition manual is a real concept translated
into programming.

Think about a content creator that always edits
his videos in the same way (him or his paid
editor). The editor has a bank of memes, sounds,
images, and tries to do the same in all the videos
he edits. This is using your edition guideline.
Applying your edition manual. And that's the idea
we want to build here.

When you are editing, you put some images when some
concept is mentioned, or maybe a sound is played
when you say a joke. Or, simply, a censorship sound
is added when a bad word is said.

All this can be automated by writting this edition
manual and telling the software how and when you
want the clip to be edited with any resource you
want. You maybe want an effect, a sound, a sticker
image, some specific video clip played overlay
during X seconds. All this is what you can do here.

Here is an example of an Edition Manual based on
football topic:

'terms': {
    'Lionel Messi': {
        'mode': 'exact',
        'context': 'generic',
        'enhancements': [
            {
                'type': 'sticker',
                'keywords': 'lionel messi portrait',
                'url': '',
                'filename': '',
                'mode': 'overlay'
            }
        ]
    },
    'Cristiano Ronaldo': {
        'mode': 'exact',
        'context': 'generic',
        'enhancements': [
            {
                'type': 'sticker',
                'keywords': 'cristiano ronaldo portrait',
                'url': '',
                'filename': '',
                'mode': 'overlay'
            }
        ]
    }
}
"""
from yta_core.edition_manual.term import EditionManualTerm
from yta_audio.voice.transcription.objects import AudioTranscription
from yta_general_utils.file.reader import FileReader
from yta_general_utils.programming.validator.parameter import ParameterValidator


class EditionManual:
    """
    Class to encapsulate all the terms to be 
    used to edit a video in the way the user
    expects with this configuration.

    This EditionManual will hold a lof of terms
    that will be transformed into effects,
    videos, texts and any other kind of elements
    supporting the raw video according to the
    specifications of these terms.
    """
    
    terms: list[EditionManualTerm] = None
    """
    The list containing all the terms that
    belongs to this Edition Manual and are
    applicable here.
    """

    def __init__(
        self,
        terms: list[EditionManualTerm]
    ):
        # TODO: We need an ID or maybe we can use
        # the json as the identifier (as we do with
        # the projects)
        self.terms = terms

    @staticmethod
    def init_from_file(
        filename: str
    ) -> 'EditionManual':
        """
        Initialize an Edition Manual from the
        given file 'filename' that must have a
        valid structure.
        """
        ParameterValidator.validate_mandatory_string('filename', filename, do_accept_empty = False)

        file_content = FileReader.read_json(filename)
        # Check if file is valid
        if file_content.get('terms', None) is None:
            raise Exception(f'No "terms" field found in the provided json filename {filename}.')

        return EditionManual.init_from_dict(
            file_content['terms']
        )

    @staticmethod
    def init_from_dict(
        terms: dict
    ) -> 'EditionManual':
        """
        Initialize an EditionManual instance with
        each of the provided 'terms' being
        instantiated as EditionManualTerms.
        """
        return EditionManual([
            # This method will raise an Exception
            # if some term is not valid
            EditionManualTerm.init_from_dict(
                key,
                content
            )
            for key, content in terms.items()
        ])

    def apply(
        self,
        transcription: AudioTranscription
    ) -> list:
        """
        Apply this edition manual to the given
        'transcription' so we are able to find the
        terms we are looking for and the indexes
        in which those terms are found, that will
        be converted to the time moments to be
        able to place the improvements in the
        moment they are expected to be applied.
        """
        return [
            item
            for term in self.terms
            for item in term.search(transcription)
        ]


from yta_core.builder import Builder
from yta_core.builder.youtube import YoutubeDownloader
from yta_core.enums.field import EnhancementField, SegmentField
from yta_core.settings import Settings
from yta_general_utils.programming.validator.parameter import ParameterValidator
from moviepy import VideoFileClip, concatenate_videoclips
from moviepy.Clip import Clip
from typing import Union


class YoutubeVideoBuilder(Builder):
    """
    The builder of the YOUTUBE_VIDEO type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ) -> Clip:
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return YoutubeVideoBuilder.build(
            enhancement.get(EnhancementField.URL.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ) -> Clip:
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return YoutubeVideoBuilder.build(
            segment.get(SegmentField.URL.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )

    @staticmethod
    def build(
        url: str,
        duration: Union[float, int]
    ) -> Clip:
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)

        # Get the youtube video
        youtube_downloader = YoutubeDownloader()
        # This will raise an Exception if not available
        youtube_video = youtube_downloader.get_video(url)

        # TODO: We should handle different types of build
        # processes when a YOUTUBE_VIDEO is requested with
        # a 'strategy' field or something. What about
        # 'hot_moments', 'random_scenes', 'summarize' (?)
        # TODO: What if asking for an 'strategy' that is 
        # not available (?)
        STRATEGY = 'HOT_MOMENTS'

        number_of_scenes = duration // Settings.MAX_DURATION_PER_YOUTUBE_SCENE
        if (duration + Settings.MAX_DURATION_PER_YOUTUBE_SCENE) > 0:
            number_of_scenes += 1

        duration_of_scene = duration / number_of_scenes

        # TODO: Build 'number_of_scene' scenes of 'duration_of_scene'
        # seconds of duration according to the strategy
        
        # TODO: Validate if it is possible to build
        if (
            STRATEGY == 'HOT_MOMENTS' and
            not youtube_video.has_most_viewed_scenes
        ):
            raise Exception(f'The youtube video {youtube_video.id} does not have most viewed scenes.')

        if STRATEGY == 'HOT_MOMENTS':
            scenes = youtube_video.most_viewed_scenes

        # TODO: We need to build the scenes with 'duration_of_scene'
        # and 'number_of_scenes'

        # youtube_video_scenes = []
        # if youtube_video.heatmap:
        #     youtube_video_scenes = youtube_video.get_hottest_scenes(scenes_number, scene_duration)
        # else:
        #     youtube_video_scenes = youtube_video.get_scenes(scenes_number, scene_duration)

        # Now we have all scenes, subclip the youtube clip
        youtube_clip = VideoFileClip(youtube_downloader.download_this_video(youtube_video))

        # TODO: By now, as I don't have strategies to apply, I
        # am just subclipping the video in the middle
        return youtube_clip.with_subclip(youtube_video.duration - (duration / 2), youtube_video.duration + (duration / 2))

        return concatenate_videoclips([
            youtube_clip.with_subclip(scene.start_time, scene.end_time)
            for scene in scenes
        ])


from yta_core.builder import Builder
from yta_core.builder.dataclasses import ClippableImage
from yta_core.builder.stock import StockDownloader
from yta_core.builder.ai_image_builder import AIImageBuilder
from yta_core.enums.field import EnhancementField, SegmentField
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_multimedia.video.edition.resize import resize_video, ResizeMode
from yta_general_utils.logger import print_in_progress
from moviepy import concatenate_videoclips, VideoFileClip
from typing import Union


class StockBuilder(Builder):
    """
    The builder of the STOCK type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        """
        Build the video content from the information
        in the given 'enhancement' dict.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return StockBuilder.build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ):
        """
        Build the video content from the information
        in the given 'segment' dict.
        """
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return StockBuilder.build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )
    
    @classmethod
    def build(
        cls,
        keywords: str,
        duration: Union[float, int]
    ):
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)
        
        stock_downloader = StockDownloader(do_ignore_repeated = True)
        
        videos = []
        accumulated_duration = 0
        while accumulated_duration < duration:
            print_in_progress('Downloading stock video')
            # TODO: Make this force 1920x1080 resolution
            downloaded_filename = stock_downloader.download_video(
                keywords = keywords,
                # Maybe randomizing is not good because the
                # best results are the first ones
                do_randomize = True
            )

            # If no video, download stock image
            if not downloaded_filename:
                downloaded_filename = stock_downloader.download_image(
                    keywords = keywords,
                    # Maybe randomizing is not good because the
                    # best results are the first ones
                    do_randomize = True
                )

                if downloaded_filename:
                    video = VideoFileClip(ClippableImage(downloaded_filename, duration - accumulated_duration).video)
                else:
                    # No stock videos available, lets build with AI images
                    # TODO: Maybe we should improve the prompt for stock AI
                    # images and set it as a common method
                    video = AIImageBuilder.build(keywords, duration - accumulated_duration)
                    # TODO: If 'video' fails here... omg..., but can happen
            else:
                video = VideoFileClip(downloaded_filename)

            # Resize the video to fit custom scene size
            # TODO: Read these values from size constants
            video = resize_video(video, (1920, 1080), resize_mode = ResizeMode.RESIZE_KEEPING_ASPECT_RATIO)

            accumulated_duration += video.duration
            # Last clip must be cropped to fit the expected duration
            if accumulated_duration > duration:
                video = video.with_subclip(0, video.duration - (accumulated_duration - duration))
            # TODO: I'm forcing 1920, 1080 here but it must come from Pexels
            videos.append(video)

        return concatenate_videoclips(videos)


from yta_core.classes.segment_json import SegmentJson
from yta_core.enums.field import SegmentBuildingField
from yta_core.enums.type import SegmentType
from yta_core.enums.status import SegmentStatus
from yta_core.edition_manual import EditionManual
from yta_core.settings import Settings
from yta_core.builder import get_builder_class
from yta_core.database import database_handler
from yta_core.shortcodes.parser import shortcode_parser
from yta_core.audio.transcription import AudioTranscription
from yta_audio.voice.transcription import DefaultTimestampedAudioTranscriptor
from yta_audio.voice.generation import MicrosoftVoiceNarrator
from yta_audio.silences import AudioSilence
from yta_general_utils.temp import Temp
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.file.handler import FileHandler
from yta_general_utils.logger import print_completed, print_in_progress
from yta_general_utils.file.filename import get_file_extension
from moviepy import VideoFileClip, AudioFileClip, CompositeAudioClip
from bson.objectid import ObjectId
from typing import Union


class Segment:
    """
    A segment within a Project that is able to build
    itself according to its configuration and values.
    """

    project_id: ObjectId
    """
    The project this segment belongs to.
    """
    index: int
    """
    The index of the segment, which identifies its
    order within the project.
    """
    status: str
    """
    The current status of the segment.
    """
    data: SegmentJson
    """
    The data that the segment has.
    """
    transcription: Union[AudioTranscription, None]
    """
    The audio voice narration transcription.
    """
    audio_filename: str
    """
    The filename of the audio file.
    """
    video_filename: str
    """
    The filename of the video file.
    """
    full_filename: str
    """
    The filename of the full file that includes
    audio and video.
    """

    @property
    def duration(
        self
    ) -> float:
        """
        The duration of the clip, that is read from
        the 'data' dict.
        """
        return self.data.duration
    
    @duration.setter
    def duration(
        self,
        value: float
    ) -> None:
        """
        Set the 'duration' property, that is located
        in the 'data' dict.
        """
        self.data.duration = value

    @property
    def audio_narration_filename(
        self
    ) -> str:
        """
        The audio voice narration filename, that is
        read from the 'data' dict.
        """
        return self.data.audio_narration_filename
    
    @audio_narration_filename.setter
    def audio_narration_filename(
        self,
        value: str
    ) -> None:
        """
        Set the 'audio_narration_filename' property,
        that is located in the 'data' dict.
        """
        self.data.audio_narration_filename = value

    @property
    def has_audio_filename(
        self
    ) -> bool:
        """
        Check if the segment has an 'audio_filename' set.
        """
        return self.audio_filename is not None

    @property
    def audio_clip(
        self
    ) -> Union[AudioFileClip, None]:
        """
        The audio clip, from the audio filename if
        existing.
        """
        return self._audio_clip
        return (
            AudioFileClip(self.audio_filename)
            if self.has_audio_filename else
            None
        )
    
    @audio_clip.setter
    def audio_clip(
        self,
        value
    ) -> None:
        # TODO: Improve this
        self._audio_clip = value

    @property
    def has_audio_clip(
        self
    ) -> bool:
        """
        Check if the segment has an audio clip or not.
        """
        return self.audio_clip is not None
    
    @property
    def has_video_filename(
        self
    ) -> bool:
        """
        Check if the segment has a 'video_filename' set.
        """
        return self.video_filename is not None

    @property
    def video_clip(
        self
    ) -> Union[VideoFileClip, None]:
        """
        The video clip, from the video filename if
        existing.
        """
        return self._video_clip
        return (
            VideoFileClip(self.video_filename)
            if self.has_video_filename else
            None
        )

    @video_clip.setter
    def video_clip(
        self,
        value
    ) -> None:
        # TODO: Improve this
        self._video_clip = value
    
    @property
    def has_video_clip(
        self
    ) -> bool:
        """
        Check if the segment has a video clip or not.
        """
        return self.video_clip is not None
    
    @property
    def has_full_filename(
        self
    ) -> bool:
        """
        Check if the segment has a 'full_filename' set.
        """
        return self.full_filename is not None

    @property
    def full_clip(
        self
    ) -> Union[VideoFileClip, None]:
        """
        The full clip, including video and audio, from
        the full filename if existing.
        """
        return self._full_clip
        return (
            VideoFileClip(self.full_filename)
            if self.has_full_filename else
            None
        )
    
    @full_clip.setter
    def full_clip(
        self,
        value
    ) -> None:
        # TODO: Improve this
        self._full_clip = value
    
    @property
    def has_full_clip(
        self
    ) -> bool:
        """
        Check if the segment has a full clip or not.
        """
        return self.full_clip is not None

    @property
    def has_transcription(
        self
    ) -> bool:
        """
        Check if the segment has a transcription or not.
        """
        return self.transcription is not None

    def __init__(
        self,
        project_id: ObjectId,
        index: int,
        data: dict
    ):
        ParameterValidator.validate_mandatory_instance_of('project_id', project_id, ObjectId)
        ParameterValidator.validate_mandatory_int('index', index)

        self.project_id = project_id
        self.index = index
        self.status = data.get(SegmentBuildingField.STATUS.value, None)

        self.data = SegmentJson()

        # TODO: What about 'status' that is both in 'data'
        # and as a specific attribute (?)
        for key in data:
            setattr(self.data, key, data[key])

        # TODO: Use 'set_audio_filename' (...) here (?)
        self.transcription = data.get(SegmentBuildingField.TRANSCRIPTION.value, None)
        self.audio_filename = data.get(SegmentBuildingField.AUDIO_FILENAME.value, None)
        self.audio_clip = (
            AudioFileClip(self.audio_filename)
            if self.audio_filename is not None else
            None
        )
        self.video_filename = data.get(SegmentBuildingField.VIDEO_FILENAME.value, None)
        self.video_clip = (
            VideoFileClip(self.video_filename)
            if self.video_filename is not None else
            None
        )
        self.full_filename = data.get(SegmentBuildingField.FULL_FILENAME.value, None)
        self.full_clip = (
            VideoFileClip(self.full_filename)
            if self.full_filename is not None else
            None
        )

    def build(
        self
    ):
        print_completed('Build completed')
        # TODO: Do it
        from yta_core.configuration import Configuration
        from yta_core.configuration.condition.optional import DoShouldBuildNarration

        configuration = Configuration.get_configuration_by_type(self.data.type)

        if DoShouldBuildNarration.is_satisfied(configuration, self.data.as_dict):
            # 1. Step: Create voice narration
            if not self.has_audio_filename:
                self._create_narration()

            # Force duration according to the audio
            self.duration = self.audio_clip.duration

            # 2. Step: Create transcription
            if not self.has_transcription:
                # TODO: Create transcription
                self._create_transcription()

            # 3. Step: Extract user shortcodes
            # TODO: Handle this (maybe status (?))
            self._extract_user_shortcodes()

            # 4. Step: Apply edition manual shortcodes
            self._apply_edition_manual()

        # TODO: What about duration here that is not set (?)
        # TODO: Should we '.set_duration' here to persist it (?)
        self.duration = (
            self.audio_clip.duration
            if self.has_audio_clip else
            self.duration
        )

        # 5. Step: Create visual (video) content
        if not self.has_video_filename:
            print_in_progress('Building base content step 4')
            self._build_video_content()
            print_completed('Base content built in step 4')

        # This would be the real duration that could be not the
        # one in the database as this is in real time and it is
        # not updated in the database
        self.duration = self.video_clip.duration

        # 6. Step: Compound video and audio (if necessary)
        if self.audio_clip:
            # If we have another audio clip we should want to put
            # them togeter, not to replace it
            audio = self.audio_clip
            if self.video_clip.audio:
                # We have another audio, we need to put them together
                # TODO: What about both audio lenghts? Are always
                # the same (?)
                audio = CompositeAudioClip([
                    self.video_clip.audio,
                    audio
                ])
                
            # TODO: Should we replace self.audio_clip (?)
            self.video_clip = self.video_clip.with_audio(audio)
        elif not self.video_clip.audio:
            # Premade and other type of video_clip can have audio
            # by themselves, but another type of videos cannot, so
            # we need to put silence audio to have an audio track
            # or we will have issues when concatenating with ffmpeg
            # in those clips without audio
            self.video_clip = self.video_clip.with_audio(
                AudioSilence.create(self.video_clip.duration)
            )

        # 7. Step: Update enhancements duration
        # TODO: By now I'm omitting this part
        #self.build_step_5_update_enhancements_duration()
        # 8. Step: Build and apply enhancements
        # TODO: By now I'm omitting this part
        #self.build_step_6_build_and_apply_enhancements()

        # 9. Step: Build the final video
        self.full_clip = self.video_clip
        filename = self._create_segment_file('definitivo.mp4')
        self.full_clip.write_videofile(filename)
        self._set_full_filename(filename)

        self._set_as_finished()

    def _set_audio_filename(
        self,
        audio_filename: Union[str, None]
    ):
        """
        Set the 'audio_filename' parameter and update
        it in the database.
        """
        ParameterValidator.validate_string('audio_filename', audio_filename, do_accept_empty = False)

        self.audio_filename = audio_filename
        self.audio_clip = AudioFileClip(self.audio_filename)

        if audio_filename is not None:
            database_handler.update_project_segment_field(
                self.project_id,
                self.index,
                SegmentBuildingField.AUDIO_FILENAME.value,
                self.audio_filename
            )

    def _set_video_filename(
        self,
        video_filename: Union[str, None]
    ):
        """
        Set the 'video_filename' parameter and update
        it in the database.
        """
        ParameterValidator.validate_string('video_filename', video_filename, do_accept_empty = False)

        self.video_filename = video_filename
        self.video_clip = VideoFileClip(self.video_filename)

        if video_filename is not None:
            database_handler.update_project_segment_field(
                self.project_id,
                self.index,
                SegmentBuildingField.VIDEO_FILENAME.value,
                self.video_filename
            )

    def _set_full_filename(
        self,
        full_filename: Union[str, None]
    ):
        """
        Set the 'video_filename' parameter and update
        it in the database.
        """
        ParameterValidator.validate_string('full_filename', full_filename, do_accept_empty = False)

        self.full_filename = full_filename
        self.full_clip = VideoFileClip(self.full_filename)

        if full_filename is not None:
            database_handler.update_project_segment_field(
                self.project_id,
                self.index,
                SegmentBuildingField.FULL_FILENAME.value,
                self.full_filename
            )

    def _set_transcription(
        self,
        transcription: AudioTranscription
    ):
        """
        Set the transcription in the instance and also in
        the database.
        """
        ParameterValidator.validate_instance_of('transcription', transcription, 'AudioTranscription')

        # We have AudioTranscription from 'yta_audio', but
        # we want the one from 'yta_core'
        transcription = AudioTranscription(transcription.words)

        self.transcription = transcription

        if transcription is not None:
            database_handler.update_project_segment_field(
                self.project_id,
                self.index,
                SegmentBuildingField.TRANSCRIPTION.value,
                # TODO: Turn AudioTranscription (raw) into his wrapper
                self.transcription.for_mongo
            )

    def _set_shortcodes(
        self,
        shortcodes
    ):
        #  TODO: What type are these ones (?)
        self.shortcodes = shortcodes

        if shortcodes is not None:
            database_handler.update_project_segment_field(
                self.project_id,
                self.index,
                SegmentBuildingField.SHORTCODES.value,
                self.shortcodes
            )

    def _set_as_finished(
        self
    ):
        """
        Set the segment as finished (building has been
        completed).
        """
        database_handler.update_project_segment_status(
            self.project_id,
            self.index,
            SegmentStatus.FINISHED.value,
        )

    def _create_segment_file(
        self,
        filename: str
    ):
        """
        Create a filename within the definitive segments
        folder to keep the generated file locally and
        recover it later if something goes wrong.

        The definitive filename will be built using the 
        provided 'filename' and adding some more
        information in the name like the current segment
        index.
        """
        ParameterValidator.validate_mandatory_string('filename', filename, do_accept_empty = False)

        return f'{Settings.DEFAULT_SEGMENT_PARTS_FOLDER}/segment_{self.index}_{Temp.get_filename(filename)}'
    
    def _create_narration(
        self
    ):
        """
        Create the audio narration (if needed) by using
        AI audio narration, or copies the original audio
        narration to the definitive filename.
        """
        if self.audio_narration_filename:
            segment_part_filename = self._create_segment_file(f'narration.{get_file_extension(self.audio_narration_filename)}')
            FileHandler.copy_file(self.audio_narration_filename, segment_part_filename)
            print_completed('Original voice narration file copied to segment parts folder')
            self.audio_narration_filename = segment_part_filename
            self._set_audio_filename(segment_part_filename)
        else:
            segment_part_filename = self._create_segment_file('narration.wav')
            # TODO: Voice parameter need to change
            self._set_audio_filename(
                MicrosoftVoiceNarrator.narrate(
                    # TODO: How to get this field properly (?)
                    self.data.text_to_narrate_sanitized_without_shortcodes,
                    output_filename = segment_part_filename
                )
            )
            print_completed('Voice narration created successfully')

    def _create_transcription(
        self
    ):
        """
        Create the transcription of the audio narration.

        Creates the transcription of the generated audio narration
        that would be stored in 'self.audio_filename'.
        
        This method returns a words array containing, for each word,
        a 'text', 'start' and 'end' field to be able to use the 
        transcription timestamps.
        """
        self._set_transcription(
            DefaultTimestampedAudioTranscriptor.transcribe(
                self.audio_filename,
                initial_prompt = self.data.text_to_narrate_sanitized_without_shortcodes
            )
        )

    def _extract_user_shortcodes(
        self
    ):
        shortcode_parser.parse(self.data.text_to_narrate)
        #self.shortcodes = shortcode_parser.shortcodes

        self._set_shortcodes(shortcode_parser.shortcodes)

        # TODO: Turn shortcodes into enhancements
        # TODO: I'm doing nothing with these enhancements
        # TODO: Maybe store in database (?)
        # TODO: Maybe I should manage this as 'user-enhancements'
        # enhancements = [
        #     shortcode.to_enhancement_element(self.transcription)
        #     for shortcode in self.shortcodes
        # ]

    def _apply_edition_manual(
        self
    ):
        # TODO: I need to dynamically get the edition manual from somewhere
        # By now I'm forcing it
        test_edition_manual = 'C:/Users/dania/Desktop/PROYECTOS/youtube-autonomous/youtube_autonomous/segments/enhancement/edition_manual/example.json'
        edition_manual = EditionManual.init_from_file(test_edition_manual)
        dict_enhancements_found = edition_manual.apply(self.transcription)
        print(dict_enhancements_found)
        # Turn dict enhancements found into Enhancement objects
        for index, dict_enhancement_found in enumerate(dict_enhancements_found):
            # TODO: What do we do with 'index' here (?)
            index = 100 + index
            self.enhancements.append(Enhancement(self.project_id, self.index, self, index, dict_enhancement_found))
        # TODO: Some enhancements could be incompatible due to collisions
        # or things like that

        # TODO: What about this enhancements (?)

    def _build_video_content(
        self
    ):
        builder = get_builder_class(SegmentType.to_enum(self.data.type))()
        # TODO: Is the 'data' enough (?)
        video_clip = builder.build_from_segment(self.data.as_dict)
        filename = self._create_segment_file('video.mp4')
        video_clip.write_videofile(filename)
        self._set_video_filename(filename)



from yta_core.builder import Builder
from yta_core.builder.stock_builder import StockBuilder
from yta_core.enums.field import EnhancementField, SegmentField
from yta_core.builder.youtube import YoutubeDownloader
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.logger import print_in_progress
from moviepy import VideoFileClip, concatenate_videoclips
from typing import Union


class CustomStockBuilder(Builder):
    """
    The builder of the AI_VIDEO type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        """
        Build the video content from the information
        in the given 'enhancement' dict.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return CustomStockBuilder.build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ):
        """
        Build the video content from the information
        in the given 'segment' dict.
        """
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return CustomStockBuilder.build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )
    
    @classmethod
    def build(
        cls,
        keywords: str,
        duration: Union[float, int]
    ):
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)
        
        youtube_downloader = YoutubeDownloader()

        # TODO: This way of building the final video has
        # to be reviewed
        videos = []
        do_use_youtube = True # to stop searching in Youtube if no videos available
        accumulated_duration = 0
        while accumulated_duration < duration:
            downloaded_video = None
            if do_use_youtube:
                # We try to download if from Youtube
                print_in_progress('Downloading youtube stock video')
                youtube_stock_video = youtube_downloader.get_stock_video(keywords)
                if youtube_stock_video is not None:
                    downloaded_video = youtube_stock_video.download()
                    if downloaded_video is not None:
                        youtube_downloader.ignore_id(youtube_stock_video.id)

            if downloaded_video is None:
                # Not found or not searching on Youtube, so build 'stock'
                print_in_progress('Downloading stock video (youtube not found)')
                do_use_youtube = False
                video = StockBuilder.build(keywords, duration)
            else:
                video = VideoFileClip(downloaded_video.output_filename)

            accumulated_duration += video.duration

            if accumulated_duration > duration:
                end = video.duration - (accumulated_duration - duration)
                start = 0
                if (
                    youtube_stock_video and
                    youtube_stock_video.key_moment != 0 and
                    youtube_stock_video.key_moment is not None
                ):
                    # Ok, lets use that key moment as the center of our video
                    start = youtube_stock_video.key_moment - (end / 2)
                    end = youtube_stock_video.key_moment + (end / 2)
                    if start < 0:
                        end += abs(0 - start)
                        start = 0
                    if end > video.duration:
                        start -= abs(end - video.duration)
                        end = video.duration
                video = video.with_subclip(start, end)

            videos.append(video)

        return concatenate_videoclips(videos)


from yta_core.builder import Builder
from yta_core.builder.youtube import YoutubeDownloader
from yta_core.enums.field import EnhancementField, SegmentField
from yta_general_utils.programming.validator.parameter import ParameterValidator
from moviepy import VideoFileClip
from typing import Union


__all__ = [
    'MemeBuilder'
]

class MemeBuilder(Builder):
    """
    The builder of the MEME type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ) -> VideoFileClip:
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return MemeBuilder.build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ) -> VideoFileClip:
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return MemeBuilder.build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )

    @classmethod
    def build(
        cls,
        keywords: str,
        duration: Union[float, int]
    ) -> VideoFileClip:
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration)

        youtube_downloader = YoutubeDownloader()

        youtube_downloader.deactivate_ignore_repeated()
        temp_filename = youtube_downloader.download_meme_video(keywords, True, True)
        youtube_downloader.activate_ignore_repeated()

        # TODO: Look for a better strategy (?)
        if not temp_filename:
            raise Exception(f'No meme found with the given "keywords": {keywords}.')
        
        video = VideoFileClip(temp_filename)

        return (
            video.with_subclip(0, duration)
            if duration < video.duration else
            video
        )


from yta_core.builder import Builder
from yta_core.builder.enums import TextPremade
from yta_core.builder.utils import enum_name_to_class
from yta_core.enums.field import EnhancementField, SegmentField
from yta_general_utils.programming.validator.parameter import ParameterValidator
from moviepy import VideoFileClip


__all__ = [
    'TextBuilder'
]

class TextBuilder(Builder):
    """
    The builder of the TEXT type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ) -> VideoFileClip:
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return TextBuilder.build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ) -> VideoFileClip:
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return TextBuilder.build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment
        )
    
    @staticmethod
    def build(
        keywords: str,
        element: dict
    ) -> VideoFileClip:
        """
        For internal use only.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_dict('element', element)

        method = enum_name_to_class(keywords, TextPremade).generate
        parameters = TextBuilder.get_parameters_from_method(method, element)

        # TODO: Maybe 'mask' to set background as transparent if available (?)
        return VideoFileClip(
            method(**parameters)
        )
        

# TODO: Remove this code below when confirmed that is not used
# def _build_custom_from_text_premade_name(
#     text_premade_name: str,
#     **parameters
# ) -> VideoFileClip:
#     """
#     This method instantiates the 'text_animation_class' Manim
#     text animation class and uses the provided 'parameters' to
#     build the text animation. The provided 'parameters' must 
#     fit the ones requested by the provided class 'generate'
#     method.
#     """
#     ParameterValidator.validate_mandatory_string('text_premade_name', text_premade_name, do_accept_empty = False)

#     # TODO: Maybe 'mask' to set background as transparent if available (?)
#     return VideoFileClip(
#         _text_premade_name_to_class(text_premade_name)().generate(**parameters)
#     )


from yta_core.classes.segment import Segment
from yta_core.database import database_handler
from yta_core.enums.field import ProjectBuildingField
from yta_core.enums.status import ProjectStatus, SegmentStatus
from yta_multimedia.video.utils.ffmpeg_handler import FfmpegHandler
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.parameter import ParameterValidator
from bson.objectid import ObjectId
from typing import Union


class Project:
    """
    Class that represents a whole video Project, containing
    different segments that are used to build consecutively
    to end being a whole video that is this project video.
    """

    id: ObjectId = None
    """
    The mongo ObjectId that identifies this project in the
    database.
    """
    _status: ProjectStatus = None
    """
    The current project status that allows us to know if
    this project has started or not, or even if it has been
    finished.
    """
    _segments: list[Segment] = None
    """
    The array that contains this project segments that are
    used to build the whole project video.
    """
    _do_update_database: bool = True
    """
    Internal variable to know if we should update the
    database value.

    _This parameter is not manually set by the user._
    """

    @property
    def unfinished_segments(
        self
    ) -> list[Segment]:
        """
        Get all the segments that have not been built yet
        (they are unfinished) from this project.
        """
        return [
            segment
            for segment in self.segments
            if segment.status != SegmentStatus.FINISHED.value
        ]
    
    @property
    def status(
        self
    ):
        """
        The current project status that allows us to know
        if this project has started or not, or even if it
        has been finished.
        """
        return self._status

    @status.setter
    def status(
        self,
        status: Union[ProjectStatus, str] = ProjectStatus.TO_START
    ):
        """
        Updates the 'status' property and also updates it
        in the database if it must be done.
        """
        status = ProjectStatus.to_enum(status)

        self._status = status.value
        
        if self._do_update_database:
            database_handler.update_project_field(
                self.id,
                ProjectBuildingField.STATUS.value,
                status.value
            )

    @property
    def segments(
        self
    ):
        """
        The array that contains this project segments that
        are used to build the whole project video.
        """
        return self._segments

    @segments.setter
    def segments(
        self,
        segments: list[Segment]
    ):
        """
        Update the 'segments' property with the provided
        'segments' parameter. This methid will check that
        any of the provided segments are SegmentJson instances.
        """
        ParameterValidator.validate_mandatory_list_of_these_instances('segments', segments, Segment)

        self._segments = segments

    def __init__(
        self,
        id: Union[str, ObjectId]
    ):
        ParameterValidator.validate_mandatory_instance_of('id', id, [str, ObjectId])

        self.id = (
            id
            if PythonValidator.is_instance(id, ObjectId) else
            ObjectId(id)
        )
        self.refresh()

    def refresh(
        self
    ):
        """
        Refresh the Project data by reading it from the
        database.
        """
        project = database_handler.get_project_by_id(self.id)

        if project is None:
            raise Exception(f'There is no project in the database with the provided "{str(self.id)}" id.')

        self._do_update_database = False

        self.status = project.json['status']
        self.segments = [
            Segment(self.id, index, segment)
            for index, segment in enumerate(project.json['segments'])
        ]

        self._do_update_database = True
        
    def build(
        self,
        output_filename: str
    ) -> bool:
        """
        Build the Project by building each segment and
        putting them together, one after another, in 
        order. This method will only build the segments
        that have not been build yet.

        The result will be stored as 'output_filename'.
        """
        # I make, by now, 'output_filename' mandatory for this purpose
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)

        self.status = ProjectStatus.IN_PROGRESS

        for segment in self.unfinished_segments:
            segment.build()

        self.refresh()

        unfinished_segments_len = len(self.unfinished_segments)
        if unfinished_segments_len > 0:
            raise Exception(f'There are {str(unfinished_segments_len)} segments that have not been completely built (unfinished).')
        
        # I put them together in a whole project clip
        abspath = get_project_abspath()
        full_abspath_filenames = [
            f'{abspath}{segment.full_filename}'
            for segment in self.segments
        ]
        output_abspath = output_filename
        FfmpegHandler.concatenate_videos(full_abspath_filenames, output_abspath)

        self.status = ProjectStatus.FINISHED

        return True


from yta_core.builder.enums import Premade, TextPremade, EffectPremade
from typing import Union


def enum_name_to_class(
    premade_name: str,
    enum: Union[Premade, TextPremade, EffectPremade]
):
    """
    Turn the provided 'premade_name' to its corresponding
    premade Enum and obtain the class that can be used to
    build the video.

    This method will return None if the 'premade_name' is
    not valid, or the class if it is.
    """
    from yta_general_utils.programming.validator import PythonValidator
    print(PythonValidator.is_instance(enum, TextPremade))
    # if enum is None:
    #     raise Exception('The "enum" parameter provided is not a valid premade Enum class.')
    enum = enum.get_instance_if_name_is_accepted(premade_name, do_ignore_case = True)

    if enum is None:
        raise Exception(f'The provided premade name "{premade_name}" is not valid.')
        # raise Exception(f'The provided premade name "{premade_name}" is not valid. The valid ones are: {enum.get_all_names_as_str()}')
    
    return enum.value


from yta_audio.voice.transcription.objects import AudioTranscription as BaseAudioTranscription, AudioTranscriptionWord as BaseAudioTranscriptionWord, AudioTranscriptionWordTimestamp as BaseAudioTranscriptionWordTimestamp
from yta_general_utils.programming.validator.parameter import ParameterValidator


class AudioTranscriptionWord(BaseAudioTranscriptionWord):
    """
    Wrapper to include MongoDB serialization.
    """

    @staticmethod
    def from_mongo(
        object: dict
    ) -> 'BaseAudioTranscriptionWord':
        """
        Get an object from MongoDB and rebuild it
        as a python instance.
        """
        ParameterValidator.validate_mandatory_dict('object', object)
        ParameterValidator.validate_dict_has_keys('object', object, ['word', 'start', 'end', 'confidence'])
        
        return BaseAudioTranscriptionWord(
            object.word,
            BaseAudioTranscriptionWordTimestamp(
                object.start,
                object.end
            ),
            object.confidence
        )
    
    @property
    def for_mongo(
        self
    ) -> str:
        """
        Serialization to be stored in MongoDB.
        """
        return self.as_dict
    
class AudioTranscription(BaseAudioTranscription):
    """
    Wrapper to include MongoDB serialization.
    """

    @staticmethod
    def from_mongo(
        object: list[dict]
    ) -> 'AudioTranscriptionWord':
        """
        Get an object from MongoDB and rebuild it
        as a python instance.

        The 'object' must be an array of words.
        """
        ParameterValidator.validate_mandatory_list_of_these_instances('object', object, dict)

        return AudioTranscription([
            AudioTranscriptionWord.from_mongo(item)
            for item in object
        ])
    
    @property
    def for_mongo(
        self
    ) -> str:
        """
        Serialization to be stored in MongoDB.
        """
        return self.as_dict['words']
        


"""
Conditions that are mandatory according to the 
component configuration. These conditions, if
applicable, will raise an exception if not met.
"""
from yta_core.enums.field import _Field
from yta_general_utils.programming.validator.parameter import ParameterValidator
from abc import ABC, abstractmethod


class MandatoryCondition(ABC):
    """
    Class to representa a component configuration 
    mandatory condition that will raise an Exception
    if not met or will be applied if met.
    """

    @staticmethod
    @abstractmethod
    def is_satisfied(
        component: dict
    ):
        pass

class DoNeedKeywords(MandatoryCondition):
    """
    Check if the 'do_need_keywords' mandatory condition
    is satisfied or not.
    """

    _attribute: str = 'do_need_keywords'

    @staticmethod
    def is_satisfied(
        component: dict
    ) -> bool:
        """
        Check if the 'do_need_keywords' mandatory condition
        is satisfied or not.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        return component.get(_Field.KEYWORDS.value, None) 
    
class DoNeedFilenameOrUrl(MandatoryCondition):
    """
    Check if the 'do_need_filename_or_url' mandatory
    condition is satisfied or not.
    """

    _attribute: str = 'do_need_filename_or_url'

    @staticmethod
    def is_satisfied(
        component: dict
    ) -> bool:
        """
        Check if the 'do_need_filename_or_url' mandatory
        condition is satisfied or not.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        # TODO: Should this be different from '' (?)
        return (
            component.get(_Field.FILENAME.value, None)  or
            component.get(_Field.URL.value, None) 
        )
    
class DoNeedText(MandatoryCondition):
    """
    Check if the 'do_need_text' mandatory
    condition is satisfied or not.
    """

    _attribute: str = 'do_need_text'

    @staticmethod
    def is_satisfied(
        component: dict
    ) -> bool:
        """
        Check if the 'do_need_text' mandatory
        condition is satisfied or not.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        # TODO: Should this be different from '' (?)
        return component.get(_Field.TEXT.value, None) 
    
class DoNeedSpecificDuration(MandatoryCondition):
    """
    Check if the 'do_need_specific_duration' mandatory
    condition is satisfied or not.
    """

    _attribute: str = 'do_need_specific_duration'

    @staticmethod
    def is_satisfied(
        component: dict
    ) -> bool:
        """
        Check if the 'do_need_specific_duration' mandatory
        condition is satisfied or not.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        return (
            component.get(_Field.DURATION.value, None)  or
            DoNeedNarration.is_satisfied(component)
        )
    
class DoNeedNarration(MandatoryCondition):
    """
    Check if the 'do_need_narration' mandatory
    condition is satisfied or not.
    """

    _attribute: str = 'do_need_narration'

    @staticmethod
    def is_satisfied(
        component: dict
    ) -> bool:
        """
        Check if the 'do_need_narration' mandatory
        condition is satisfied or not.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        return (
            component.get(_Field.AUDIO_NARRATION_FILENAME.value, None) or
            (
                component.get(_Field.VOICE.value, None) and
                component.get(_Field.TEXT_TO_NARRATE.value, None)
            )
        )


"""
This module includes a lot of enum classes to
include a single source of truth and avoid
having names and variables hardcoded in the
project.

There are some private Enum classes (they start
with '_') to reuse the names in the other Enum
classes that are expected to be used in the 
different modules of the app.
"""



from setuptools import setup, find_packages


VERSION = '0.0.10'
DESCRIPTION = 'Youtube Autonomous Core'
LONG_DESCRIPTION = 'Youtube Autonomous Core - where the magic happens.'

setup(
    name = "yta_core", 
    version = VERSION,
    author = "Daniel Alcalá",
    author_email = "<danielalcalavalera@gmail.com>",
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    packages = find_packages(),
    install_requires = [
        'yta_general_utils'
    ],
    
    keywords = [
        'youtube autonomous core',
    ],
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Field(Enum):
    """
    Enum class to represent the names of the
    fields we are able to handle in the app.

    This has been built to have all the names
    centralized in one place and to reuse in
    any app part in which it is needed.

    For internal use only. Use a more specific
    Enum class to interact with the different
    field names.
    """

    STATUS = 'status'
    """
    The status of an element, that will tell us
    if it is being constructed, waiting for it,
    etc.
    """
    SEGMENTS = 'segments'
    """
    The segments that a project includes and have
    to be built.
    """
    TYPE = 'type'
    """
    The type of some specific element. See the 
    specific type Enum classes.
    """
    TEXT = 'text'
    """
    The text that has to be written in some 
    segment with text content.
    """
    TEXT_TO_NARRATE = 'text_to_narrate'
    """
    The text that has to be used for a voice
    narration. This text can include shortcodes
    because it will be narrated and it is 
    possible to add some modifications (the ones
    that the shortcodes indicate) according to
    the moment in which some words are said.
    """
    DURATION = 'duration'
    """
    The length of time that the element being
    referenced will be displayed or occurring.
    """
    KEYWORDS = 'keywords'
    """
    The terms or sentence to make a search on
    Youtube, to create the specific premade, etc.
    Keywords to identify the expected behaviour.
    """
    URL = 'url'
    """
    The url from which we can obtain a resource
    that has to be included as part (or the
    whole) content of the element.
    """
    FILENAME = 'filename'
    """
    The local filename of the file that has to
    be included as part (or the whole) content 
    of the element.
    """
    VOICE = 'voice'
    """
    The voice that has to be used to create the
    voice narration.
    """
    AUDIO_NARRATION_FILENAME = 'audio_narration_filename'
    """
    The filename of an audio that has to be 
    included as the voice narration (so it will
    not be created).
    """
    MUSIC = 'music'
    """
    The filename of a music that has to be included
    as an audio part for the content.
    """
    ENHANCEMENTS = 'enhancements'
    """
    The enhancements that have to be built for the
    content.
    """
    START = 'start'
    """
    The time moment in which the content has to start.
    """
    MODE = 'mode'
    """
    The mode in which the content has to be included.
    """
    EXTRA_PARAMS = 'extra_params'
    """
    The extra parameters we need to build some of the
    content.
    """
    # Building fields below
    TRANSCRIPTION = 'transcription'
    """
    The transcription of the audio narration if 
    existing.
    """
    SHORTCODES = 'shortcodes'
    """
    The shortcodes found in an audio narration.
    """
    AUDIO_FILENAME = 'audio_filename'
    """
    The filename of the audio that has been 
    generated by the system and has to be used
    as the audio content part.
    """
    AUDIO_CLIP = 'audio_clip'
    """
    The clip generated from the 'audio_filename'.
    """
    VIDEO_FILENAME = 'video_filename'
    """
    The filename of the video part that has 
    been generated by the system and has to be
    used as the video content part.
    """
    VIDEO_CLIP = 'video_clip'
    """
    The clip generated from the 'video_filename'.
    """
    FULL_FILENAME = 'full_filename'
    """
    The filename of the video and audio part put
    together, generated by the system, and that
    is the whole segment content with the effects
    and everything appliedo n it.
    """
    FULL_CLIP = 'full_clip'
    """
    The clip generated from the 'full_filename'.
    """

class ProjectField(Enum):
    """
    Enum class to represent the names of the
    fields related to a Project.
    """

    STATUS = _Field.STATUS.value
    """
    The status of an element, that will tell us
    if it is being constructed, waiting for it,
    etc.

    See ProjectStatus Enum class.
    """
    SEGMENTS = _Field.SEGMENTS.value
    """
    The segments that a project includes and have
    to be built.
    """

class SegmentField(Enum):
    """
    Enum class to represent the names of the
    fields related to a Segment (which belongs
    to a Project).
    """

    TYPE = _Field.TYPE.value
    """
    The type of a segment, that determins what
    kind of content will be used to fulfill that
    segment.
    """
    TEXT = _Field.TEXT.value
    """
    The text that has to be written in some 
    segment with text content.
    """
    TEXT_TO_NARRATE = _Field.TEXT_TO_NARRATE.value
    """
    The text that has to be used for a voice
    narration. This text can include shortcodes
    because it will be narrated and it is 
    possible to add some modifications (the ones
    that the shortcodes indicate) according to
    the moment in which some words are said.
    """
    DURATION = _Field.DURATION.value
    """
    The length of time that the segment being
    referenced will be displayed or occurring.
    """
    KEYWORDS = _Field.KEYWORDS.value
    """
    The terms or sentence to make a search on
    Youtube, to create the specific premade, etc.
    Keywords to identify the expected behaviour.
    """
    URL = _Field.URL.value
    """
    The url from which we can obtain a resource
    that has to be included as part (or the
    whole) content of the element.
    """
    FILENAME = _Field.FILENAME.value
    """
    The local filename of the file that has to
    be included as part (or the whole) content 
    of the element.
    """
    VOICE = _Field.VOICE.value
    """
    The voice that has to be used to create the
    voice narration.
    """
    AUDIO_NARRATION_FILENAME = _Field.AUDIO_NARRATION_FILENAME.value
    """
    The filename of an audio that has to be 
    included as the voice narration (so it will
    not be created).
    """
    MUSIC = _Field.MUSIC.value
    """
    The filename of a music that has to be included
    as an audio part for the content.
    """
    ENHANCEMENTS = _Field.ENHANCEMENTS.value
    """
    The enhancements that have to be built for the
    content.
    """
    EXTRA_PARAMS = _Field.EXTRA_PARAMS.value
    """
    The extra parameters we need to build some of the
    content.
    """

class EnhancementField(Enum):
    """
    Enum class to represent the names of the
    fields related to an Enhancement (which
    belongs to a Project Segment).
    """

    TYPE = _Field.TYPE.value
    """
    The type of an Enhancement, that determins
    what kind of content will be used to fulfill
    that segment.
    """
    KEYWORDS = _Field.KEYWORDS.value
    """
    The terms or sentence to make a search on
    Youtube, to create the specific premade, etc.
    Keywords to identify the expected behaviour.
    """
    URL = _Field.URL.value
    """
    The url from which we can obtain a resource
    that has to be included as part (or the
    whole) content of the element.
    """
    FILENAME = _Field.FILENAME.value
    """
    The local filename of the file that has to
    be included as part (or the whole) content 
    of the element.
    """
    TEXT_TO_NARRATE = _Field.TEXT_TO_NARRATE.value
    """
    The text that has to be used for a voice
    narration. This text can include shortcodes
    because it will be narrated and it is 
    possible to add some modifications (the ones
    that the shortcodes indicate) according to
    the moment in which some words are said.
    """
    VOICE = _Field.VOICE.value
    """
    The voice that has to be used to create the
    voice narration.
    """
    TEXT = _Field.TEXT.value
    """
    The text that has to be written in some 
    segment with text content.
    """
    DURATION = _Field.DURATION.value
    """
    The length of time that the segment being
    referenced will be displayed or occurring.
    """
    AUDIO_NARRATION_FILENAME = _Field.AUDIO_NARRATION_FILENAME.value
    """
    The filename of an audio that has to be 
    included as the voice narration (so it will
    not be created).
    """
    MUSIC = _Field.MUSIC.value
    """
    The filename of a music that has to be included
    as an audio part for the content.
    """
    ENHANCEMENTS = _Field.ENHANCEMENTS.value
    """
    The enhancements that have to be built for the
    content.
    """
    START = _Field.START.value
    """
    The time moment in which the content has to start.
    """
    MODE = _Field.MODE.value
    """
    The mode in which the content has to be included.
    """
    EXTRA_PARAMS = _Field.EXTRA_PARAMS.value
    """
    The extra parameters we need to build some of the
    content.
    """

class ShortcodeField(Enum):
    """
    Enum class to represent the names of the
    fields related to a Shortcode (which
    was detected within a Segment or Enhancement
    text).
    """

    TYPE = _Field.TYPE.value
    """
    The type of a Shortcode, that determins
    what kind of content will be used to fulfill
    that segment.
    """
    KEYWORDS = _Field.KEYWORDS.value
    """
    The terms or sentence to make a search on
    Youtube, to create the specific premade, etc.
    Keywords to identify the expected behaviour.
    """
    URL = _Field.URL.value
    """
    The url from which we can obtain a resource
    that has to be included as part (or the
    whole) content of the element.
    """
    FILENAME = _Field.FILENAME.value
    """
    The local filename of the file that has to
    be included as part (or the whole) content 
    of the element.
    """
    TEXT_TO_NARRATE = _Field.TEXT_TO_NARRATE.value
    """
    The text that has to be used for a voice
    narration. This text can include shortcodes
    because it will be narrated and it is 
    possible to add some modifications (the ones
    that the shortcodes indicate) according to
    the moment in which some words are said.
    """
    VOICE = _Field.VOICE.value
    """
    The voice that has to be used to create the
    voice narration.
    """
    TEXT = _Field.TEXT.value
    """
    The text that has to be written in some 
    segment with text content.
    """
    DURATION = _Field.DURATION.value
    """
    The length of time that the segment being
    referenced will be displayed or occurring.
    """
    AUDIO_NARRATION_FILENAME = _Field.AUDIO_NARRATION_FILENAME.value
    """
    The filename of an audio that has to be 
    included as the voice narration (so it will
    not be created).
    """
    MUSIC = _Field.MUSIC.value
    """
    The filename of a music that has to be included
    as an audio part for the content.
    """
    ENHANCEMENTS = _Field.ENHANCEMENTS.value
    """
    The enhancements that have to be built for the
    content.
    """
    START = _Field.START.value
    """
    The time moment in which the content has to start.
    """
    MODE = _Field.MODE.value
    """
    The mode in which the content has to be included.
    """
    EXTRA_PARAMS = _Field.EXTRA_PARAMS.value
    """
    The extra parameters we need to build some of the
    content.
    """

class ProjectBuildingField(Enum):
    """
    The fields of a Project that are used when its
    content is being built and doesn't come from
    the original json source.
    """

    STATUS = _Field.STATUS.value
    """
    The status of the Project, that will tell us
    if it is being constructed, waiting for it,
    etc.
    """

class SegmentBuildingField(Enum):
    """
    The fields that are used when building the
    Segment and are not provided by the user in
    the initial json data.
    """

    STATUS = _Field.STATUS.value
    """
    The status of the Segment, that will tell us
    if it is being constructed, waiting for it,
    etc.
    """
    TRANSCRIPTION = _Field.TRANSCRIPTION.value
    """
    The transcription of the audio narration of
    the segment if existing.
    """
    SHORTCODES = _Field.SHORTCODES.value
    """
    The shortcodes that have been found in the
    audio narration.
    """
    AUDIO_FILENAME = _Field.AUDIO_FILENAME.value
    """
    The filename of the audio that has been 
    generated by the system and has to be used
    as the audio content part.
    """
    AUDIO_CLIP = _Field.AUDIO_CLIP.value
    """
    The clip generated from the 'audio_filename'.
    """
    VIDEO_FILENAME = _Field.VIDEO_FILENAME.value
    """
    The filename of the video part that has 
    been generated by the system and has to be
    used as the video content part.
    """
    VIDEO_CLIP = _Field.VIDEO_CLIP.value
    """
    The clip generated from the 'video_filename'.
    """
    FULL_FILENAME = _Field.FULL_FILENAME.value
    """
    The filename of the video and audio part put
    together, generated by the system, and that
    is the whole segment content with the effects
    and everything appliedo n it.
    """
    FULL_CLIP = _Field.FULL_CLIP.value
    """
    The clip generated from the 'full_filename'.
    """

class EnhancementBuildingField(Enum):
    """
    The fields that are used when building the 
    Enhancement and are not provided by the user
    in the initial json data.
    """

    STATUS = _Field.STATUS.value
    """
    The status of the Segment, that will tell us
    if it is being constructed, waiting for it,
    etc.
    """
    TRANSCRIPTION = _Field.TRANSCRIPTION.value
    """
    The transcription of the audio narration of
    the segment if existing.
    """
    AUDIO_FILENAME = _Field.AUDIO_FILENAME.value
    """
    The filename of the audio that has been 
    generated by the system and has to be used
    as the audio content part.
    """
    AUDIO_CLIP = _Field.AUDIO_CLIP.value
    """
    The clip generated from the 'audio_filename'.
    """
    VIDEO_FILENAME = _Field.VIDEO_FILENAME.value
    """
    The filename of the video part that has 
    been generated by the system and has to be
    used as the video content part.
    """
    VIDEO_CLIP = _Field.VIDEO_CLIP.value
    """
    The clip generated from the 'video_filename'.
    """
    FULL_FILENAME = _Field.FULL_FILENAME.value
    """
    The filename of the video and audio part put
    together, generated by the system, and that
    is the whole segment content with the effects
    and everything appliedo n it.
    """
    FULL_CLIP = _Field.FULL_CLIP.value
    """
    The clip generated from the 'full_filename'.
    """


"""
Module to include all the validation classes
that simplify the way we check if the provided
jsons are valid and accepted by our system or
not to build the Segment, Enhancement and/or
Shortcode components.
"""
from yta_core.configuration import Configuration
from yta_core.builder.utils import enum_name_to_class
from yta_core.builder import is_element_valid_for_method
from yta_core.builder.enums import Premade, TextPremade
from yta_core.shortcodes.parser import shortcode_parser, empty_shortcode_parser
from yta_core.enums.field import SegmentField, EnhancementField, ShortcodeField
from yta_core.enums.mode import SegmentMode, EnhancementMode, ShortcodeMode
from yta_core.enums.type import SegmentType, EnhancementType, ShortcodeType
from yta_core.enums.string_duration import SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration
from yta_core.enums.component import Component
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.validator import PythonValidator
from typing import Union


class BuilderValidator:
    """
    Class to validate everything related to segments,
    enhancements and shortcodes. A single source of
    validation.
    """

    @staticmethod
    def validate_segment_has_expected_fields(
        segment: dict
    ) -> None:
        """
        Check if the provided 'segment' dict has all the
        fields it must have as a Segment, and raises an
        Exception if not.
        """
        return BuilderValidator._validate_component_has_expected_fields(
            segment,
            Component.SEGMENT
        )
    
    @staticmethod
    def validate_enhancement_has_expected_fields(
        enhancement: dict
    ) -> None:
        """
        Check if the provided 'enhancement' dict has all the
        fields it must have as an Enhancement, and raises an
        Exception if not.
        """
        return BuilderValidator._validate_component_has_expected_fields(
            enhancement,
            Component.ENHANCEMENT
        )
    
    @staticmethod
    def validate_shortcode_has_expected_fields(
        shortcode: dict
    ) -> None:
        """
        Check if the provided 'shortcode' dict has all the
        fields it must have as a Shortcode, and raises an
        Exception if not.
        """
        return BuilderValidator._validate_component_has_expected_fields(
            shortcode,
            Component.SHORTCODE
        )

    @staticmethod
    def _validate_component_has_expected_fields(
        element: dict,
        component: Component
    ) -> None:
        """
        Check if the provided 'element' dict, that must be
        a dict representing the 'component' passed as 
        parameter, has all the fields it must have, and 
        raises an Exception if not.

        This method will detect the parameters that exist
        in the provided 'element' but are not expected to
        be on it, and the ones that are expected to be but
        are not set on it.
        """
        component = Component.to_enum(component)
        ParameterValidator.validate_mandatory_dict('element', element)

        accepted_fields = {
            Component.SEGMENT: lambda: SegmentField.get_all_values(),
            Component.ENHANCEMENT: lambda: EnhancementField.get_all_values(),
            Component.SHORTCODE: lambda: ShortcodeField.get_all_values()
        }[component]()

        accepted_fields_str = ', '.join(accepted_fields)

        unaccepted_fields = [
            key
            for key in element.keys()
            if key not in accepted_fields
        ]
        unaccepted_fields_str = ', '.join(unaccepted_fields)
        
        missing_fields = [
            field
            for field in accepted_fields
            if field not in element
        ]
        missing_fields_str = ', '.join(missing_fields)

        if missing_fields:
            raise Exception(f'The next fields are mandatory and were not found in the element: "{missing_fields_str}". The mandatory fields are: "{accepted_fields_str}".')

        if unaccepted_fields:
            raise Exception(f'The next fields are not accepted in the provided element by our system: "{unaccepted_fields_str}". The ones accepted are these: "{accepted_fields_str}".')
        
        return element
    
    # MODE below
    @staticmethod
    def validate_segment_mode_field(
        mode: Union[SegmentMode, str, None]
    ):
        """
        Validate the provided 'mode' for a Segment
        component.

        This method will raise an exception if the 
        'mode' provided is not valid.
        """
        return BuilderValidator._validate_component_mode_field(
            mode,
            Component.SEGMENT
        )
    
    @staticmethod
    def validate_enhancement_mode_field(
        mode: Union[EnhancementMode, str, None]
    ):
        """
        Validate the provided 'mode' for an Enhancement
        component.

        This method will raise an exception if the 
        'mode' provided is not valid.
        """
        return BuilderValidator._validate_component_mode_field(
            mode,
            Component.ENHANCEMENT
        )
    
    @staticmethod
    def validate_shortcode_mode_field(
        mode: Union[ShortcodeMode, str, None]
    ):
        """
        Validate the provided 'mode' for a Shortcode
        component.

        This method will raise an exception if the 
        'mode' provided is not valid.
        """
        return BuilderValidator._validate_component_mode_field(
            mode,
            Component.SHORTCODE
        )

    @staticmethod
    def _validate_component_mode_field(
        mode: Union[SegmentMode, EnhancementMode, ShortcodeMode, str, None],
        component: Component
    ):
        """
        Validate the provided 'mode' for the given
        'component'. The mode should be a SegmentMode,
        EnhancementMode or ShortcodeMode, or a string
        that fits one of these 3 enum classes.

        This method will raise an exception if the 
        'mode' provided is not valid for the given
        'component'.
        """
        component = Component.to_enum(component)

        # TODO: Do we accept 'None' value (?)
        return component.get_mode(mode)
    
    # MODE FOR TYPE below
    @staticmethod
    def validate_segment_mode_for_type(
        mode: Union[SegmentMode, str, None],
        type: Union[SegmentType, str]
    ):
        """
        Validate if the provided 'mode' is accepted by
        the also given 'type' for a Segment.
        """
        return BuilderValidator._validate_component_mode_for_type(
            mode,
            type,
            Component.SEGMENT
        )
    
    @staticmethod
    def validate_enhancement_mode_for_type(
        mode: Union[EnhancementMode, str, None],
        type: Union[EnhancementType, str]
    ):
        """
        Validate if the provided 'mode' is accepted by
        the also given 'type' for an Enhancement.
        """
        return BuilderValidator._validate_component_mode_for_type(
            mode,
            type,
            Component.ENHANCEMENT
        )
    
    @staticmethod
    def validate_shortcode_mode_for_type(
        mode: Union[ShortcodeMode, str, None],
        type: Union[ShortcodeType, str]
    ):
        """
        Validate if the provided 'mode' is accepted by
        the also given 'type' for an Shortcode.
        """
        return BuilderValidator._validate_component_mode_for_type(
            mode,
            type,
            Component.SHORTCODE
        )
    
    @staticmethod
    def _validate_component_mode_for_type(
        mode: Union[SegmentMode, EnhancementMode, ShortcodeMode, str, None],
        type: Union[SegmentType, EnhancementType, ShortcodeType, str],
        component: Component
    ):
        """
        Validate if the provided 'mode' is accepted by
        the also given 'type' for the also provided
        'component'.
        """
        component = Component.to_enum(component)

        if not component.is_mode_accepted_for_type(mode, type):
            type = (
                type
                if PythonValidator.is_string(type) else
                type.value
            )

            raise Exception(f'The "{type}" type does not accept the provided "{mode}" mode.')

    # DURATION below
    @staticmethod
    def validate_segment_duration_field(
        duration: Union[SegmentStringDuration, int, float, str, None]
    ):
        """
        Validate that the given 'duration' is valid for a
        Segment.
        """
        return BuilderValidator._validate_component_duration_field(
            duration,
            Component.SEGMENT
        )
    
    @staticmethod
    def validate_enhancement_duration_field(
        duration: Union[EnhancementStringDuration, int, float, str, None]
    ):
        """
        Validate that the given 'duration' is valid for an
        Enhancement.
        """
        return BuilderValidator._validate_component_duration_field(
            duration,
            Component.ENHANCEMENT
        )
    
    @staticmethod
    def validate_shortcode_duration_field(
        duration: Union[ShortcodeStringDuration, int, float, str, None]
    ):
        """
        Validate that the given 'duration' is valid for a
        Shortcode.
        """
        return BuilderValidator._validate_component_duration_field(
            duration,
            Component.SHORTCODE
        )

    @staticmethod
    def _validate_component_duration_field(
        duration: Union[SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration, int, float, str],
        component: Component
    ):
        """
        Validate that the provided 'duration' is valid for
        the given 'component'.
        """
        component = Component.to_enum(component)

        return component.get_duration(duration)
    
    # DURATION FOR TYPE below
    def validate_segment_duration_for_type(
        duration: Union[SegmentStringDuration, int, float, str],
        type: Union[SegmentType, str]
    ):
        """
        Validate if the provided 'duration' is accepted by
        the also given 'type' for a Segment component.
        """
        return BuilderValidator._validate_component_duration_for_type(
            duration,
            type,
            Component.SEGMENT
        )
    
    def validate_enhancement_duration_for_type(
        duration: Union[EnhancementStringDuration, int, float, str],
        type: Union[EnhancementType, str]
    ):
        """
        Validate if the provided 'duration' is accepted by
        the also given 'type' for an Enhancement component.
        """
        return BuilderValidator._validate_component_duration_for_type(
            duration,
            type,
            Component.ENHANCEMENT
        )
    
    def validate_shortcode_duration_for_type(
        duration: Union[ShortcodeStringDuration, int, float, str],
        type: Union[ShortcodeType, str]
    ):
        """
        Validate if the provided 'duration' is accepted by
        the also given 'type' for a Shortcode component.
        """
        return BuilderValidator._validate_component_duration_for_type(
            duration,
            type,
            Component.SHORTCODE
        )

    @staticmethod
    def _validate_component_duration_for_type(
        duration: Union[SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration, int, float, str],
        type: Union[SegmentType, EnhancementType, ShortcodeType, str],
        component: Component
    ):
        """
        Validate if the provided 'duration' is accepted by
        the also given 'type' for the also provided
        'component'.
        """
        component = Component.to_enum(component)

        if not component.is_duration_accepted_for_type(duration, type):
            type = (
                type
                if PythonValidator.is_string(type) else
                type.value
            )

            raise Exception(f'The "{type}" type does not accept the provided "{duration}" duration.')

    # START below
    @staticmethod
    def validate_segment_start_field(
        start: Union[int, float, str, None]
    ):
        """
        Validate that the provided 'start' is valid
        for a Segment component.
        """
        return BuilderValidator._validate_component_start_field(
            start,
            Component.SEGMENT
        )
    
    @staticmethod
    def validate_enhancement_start_field(
        start: Union[int, float, str, None]
    ):
        """
        Validate that the provided 'start' is valid
        for an Enhancement component.
        """
        return BuilderValidator._validate_component_start_field(
            start,
            Component.ENHANCEMENT
        )
    
    @staticmethod
    def validate_shortcode_start_field(
        start: Union[int, float, str, None]
    ):
        """
        Validate that the provided 'start' is valid
        for a Shortcode component.
        """
        return BuilderValidator._validate_component_start_field(
            start,
            Component.SHORTCODE
        )

    @staticmethod
    def _validate_component_start_field(
        start: Union[int, float, str, None],
        component: Component
    ):
        """
        Validate that the provided 'start' is valid
        for the given 'component'.
        """
        component = Component.to_enum(component)

        return component.get_start(start)

class SegmentJsonValidator:
    """
    Class to wrap the validation of a Segment
    that is still a raw json.
    """

    @staticmethod
    def validate(
        segment: dict
    ):
        """
        Validate a raw segment that has been read from
        a json file to check if it fits all the expected
        conditions, raising an Exception if not.
        """
        # 1. Validate that contains all the expected fields
        validate_segment_has_expected_fields(segment)

        # 2. Validate that the 'type' is valid
        validate_segment_type_is_valid(segment)

        # 3. Validate that 'text' has no shortcodes
        validate_segment_text_has_no_shortcodes(segment)

        # 4. Validate that 'text_to_narrate' doesn't have
        # invalid shortcodes
        validate_segment_text_to_narrate_has_no_invalid_shortcodes(segment)

        # 5. Validate that 'duration' is a valid string or
        # a positive numeric value
        validate_segment_duration_is_valid_string_or_positive_number(segment)

        # 6. Validate that 'duration' is FILE_DURATION for
        # a valid type
        validate_segment_duration_is_valid_for_type(segment)

        # 7. Validate if the type has the mandatory fields
        validate_segment_has_extra_params_needed(segment)

        # 8. Validate that the segment enhancements are ok
        for enhancement in segment.get(SegmentField.ENHANCEMENTS.value, []):
            EnhancementJsonValidator.validate(enhancement)

        # 9. Validate segment mandatory conditions are met
        validate_segment_mets_mandatory_conditions(segment)

class EnhancementJsonValidator:
    """
    Class to wrap the validation of a Segment
    that is still a raw json.
    """

    @staticmethod
    def validate(
        enhancement: dict
    ):
        """
        Validate a raw enhancement that has been read
        from a json file to check if it fits all the
        expected conditions, raising an Exception if
        not.
        """
        # 1. Validate that contains all the expected fields
        validate_enhancement_has_all_fields(enhancement)

        # 2. Validate that the 'type' is valid
        validate_enhancement_type_is_valid(enhancement)

        # 3. Validate that 'text' has no shortcodes
        validate_enhancement_text_has_no_shortcodes(enhancement)

        # 4. Validate that 'text_to_narrate' doesn't have
        # invalid shortcodes
        validate_enhancement_text_to_narrate_has_no_invalid_shortcodes(enhancement)

        # 5. Validate that 'duration' is a valid string or
        # a positive numeric value
        validate_enhancement_duration_is_valid_string_or_positive_number(enhancement)

        # 6. Validate that 'duration' is FILE_DURATION for
        # a valid type
        validate_enhancement_duration_is_valid_for_type(enhancement)

        # 7. Validate that 'mode' is valid
        validate_enhancement_mode_is_valid_for_type(enhancement)

        # 8. Validate all the mandatory conditions are met
        Configuration.get_configuration_by_type(
            enhancement.get(EnhancementField.TYPE, None)
        ).validate_component_mandatory_conditions(enhancement)


# Validation methods below
def validate_enhancement_has_all_fields(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' contains
    all the expected keys, which are all the ones
    available through the EnhancementField Enum
    class, and raises an Exception if not.
    """
    BuilderValidator.validate_enhancement_has_expected_fields(enhancement)
    
def validate_enhancement_type_is_valid(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has a valid
    type or raises an Exception if not.
    """
    EnhancementType.to_enum(enhancement.get(EnhancementField.TYPE.value, None))

def validate_enhancement_text_has_no_shortcodes(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has any 
    shortcode in its 'text' field and raises
    an Exception if so.
    """
    try:
        empty_shortcode_parser.parse(enhancement.get(EnhancementField.TEXT.value, ''))
    except Exception:
        raise Exception(f'The "enhancement" has some shortcodes in its "{EnhancementField.TEXT.value}" field and this is not allowed.')
    
def validate_enhancement_text_to_narrate_has_no_invalid_shortcodes(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has any
    invalid shortcode in its 'text_to_narrate'
    and raises an Exception if so.
    """
    try:
        # TODO: This has to be our general shortcode parser
        # TODO: I just faked it by now
        shortcode_parser = None
        shortcode_parser.parse(enhancement.get(EnhancementField.TEXT_TO_NARRATE.value, ''))
    except Exception:
        raise Exception(f'The "enhancement" has some invalid shortcodes in its "{EnhancementField.TEXT_TO_NARRATE.value}" field. Please, check the valid shortcodes.')
    
def validate_enhancement_duration_is_valid_string_or_positive_number(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has a 'duration'
    field that is a valid string or a positive 
    number and raises an Exception if not.
    """
    BuilderValidator.validate_enhancement_duration_field(
        enhancement.get(EnhancementField.DURATION.value, None)
    )

def validate_enhancement_duration_is_valid_for_type(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has a 'duration'
    field that is a valid string for its type or a
    positive number and raises an Exception if not.
    """
    BuilderValidator.validate_enhancement_duration_for_type(
        enhancement.get(EnhancementField.DURATION.value, None),
        enhancement.get(EnhancementField.TYPE.value, None)
    )
    
def validate_enhancement_mode_is_valid_for_type(
    enhancement: dict
):
    """
    Check if the provided 'enhancement' has a 'mode' 
    field that is valid for its type.
    """
    BuilderValidator.validate_enhancement_mode_for_type(
        enhancement.get(EnhancementField.MODE.value, None),
        enhancement.get(EnhancementField.TYPE.value, None)
    )

def validate_segment_has_expected_fields(
    segment: dict
):
    """
    Check if the provided 'segment' contains all
    the expected keys, which are all the ones
    available through the SegmentField Enum class,
    and raises an Exception if not.
    """
    BuilderValidator.validate_segment_has_expected_fields(
        segment
    )
    
def validate_segment_type_is_valid(
    segment: dict
):
    """
    Check if the provided 'segment' has a valid
    type or raises an Exception if not.
    """
    SegmentType.to_enum(segment.get(SegmentField.TYPE.value, None))

def validate_segment_text_has_no_shortcodes(
    segment: dict
):
    """
    Check if the provided 'segment' has any 
    shortcode in its 'text' field and raises
    an Exception if so.
    """
    try:
        empty_shortcode_parser.parse(segment.get(SegmentField.TEXT.value, ''))
    except Exception:
        raise Exception(f'The "segment" has some shortcodes in its "{SegmentField.TEXT.value}" field and this is not allowed.')
    
def validate_segment_text_to_narrate_has_no_invalid_shortcodes(
    segment: dict
):
    """
    Check if the provided 'segment' has any
    invalid shortcode in its 'text_to_narrate'
    and raises an Exception if so.
    """
    try:
        shortcode_parser.parse(segment.get(SegmentField.TEXT_TO_NARRATE.value, ''))
    except Exception:
        raise Exception(f'The "segment" has some invalid shortcodes in its "{SegmentField.TEXT_TO_NARRATE.value}" field. Please, check the valid shortcodes.')
    
def validate_segment_duration_is_valid_string_or_positive_number(
    segment: dict
):
    """
    Check if the provided 'segment' has a 'duration'
    field that is a valid string or a positive 
    number and raises an Exception if not.
    """
    BuilderValidator.validate_segment_duration_field(
        segment.get(SegmentField.DURATION.value, None)
    )

def validate_segment_duration_is_valid_for_type(
    segment: dict
):
    """
    Check if the provided 'segment' has a 'duration'
    field that is a valid string for its component
    type or raises an Exception if not.
    """
    BuilderValidator.validate_segment_duration_for_type(
        segment.get(SegmentField.DURATION.value, None),
        type = segment.get(SegmentField.TYPE.value, None)
    )
    
def validate_segment_has_extra_params_needed(
    segment: dict
):
    """
    Check if the provided 'segment' has the extra
    parameters that are needed according to its
    type and keywords (premades or text premades
    need extra parameters to be able to be built),
    or raises an Exception if not.
    """
    # TODO: Validate, if premade or effect, that 'extra_params' has
    # needed fields
    keywords = segment.get(SegmentField.KEYWORDS.value, None)
    if type == SegmentType.PREMADE.value:
        # TODO: This below was prepared for extra_params, I think
        # I don't have to ignore anything here... but we were
        # avoiding 'duration' because we obtain it from main fields
        if not is_element_valid_for_method(
            method = enum_name_to_class(keywords, Premade).generate,
            element = segment,
            #parameters_to_ignore = ['duration'],
            parameters_strictly_from_element = ['duration']
        ):
            # TODO: I don't tell anything about the parameters needed
            raise Exception('Some parameters are missing...')
    elif type == SegmentType.TEXT.value:
        # TODO: This below was prepared for extra_params, I think
        # I don't have to ignore anything here... but we were
        # avoiding 'text' and 'duration' because we obtain them
        # from main fields
        if not is_element_valid_for_method(
            method = enum_name_to_class(keywords, TextPremade).generate,
            element = segment,
            #parameters_to_ignore = ['output_filename', 'duration', 'text']
            parameters_to_ignore = ['output_filename'],
            parameters_strictly_from_element = ['duration', 'text']
        ):
            # TODO: I don't tell anything about the parameters needed
            raise Exception('Some parameters are missing...')
    # TODO: Validate for another types

def validate_segment_mets_mandatory_conditions(
    segment: dict
):
    """
    Check if the provided 'segment' mets all the
    mandatory conditions, that are those starting
    with 'do_' in the configuration dict and that
    have a True value, or raises an Exception if
    those mandatory conditions are not met.
    """
    Configuration.get_configuration_by_type(
        segment.get(SegmentField.TYPE.value, None)
    )().validate_component_mandatory_conditions(segment)



from yta_core.builder import Builder
from yta_core.enums.field import EnhancementField, SegmentField
from yta_general_utils.programming.validator.parameter import ParameterValidator
from moviepy import VideoFileClip
from typing import Union


__all__ = [
    'AIVideoBuilder'
]

class AIVideoBuilder(Builder):
    """
    The builder of the AI_VIDEO type.

    TODO: This has not been implemented yet.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        """
        Build the video content from the information
        in the given 'enhancement' dict.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return AIVideoBuilder._build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ):
        """
        Build the video content from the information
        in the given 'segment' dict.
        """
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return AIVideoBuilder._build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )
    
    @staticmethod
    def _build(
        keywords: str,
        duration: Union[float, int]
    ) -> VideoFileClip:
        """
        Build the video content with the given 'keywords'
        and 'duration'.
        """
        raise Exception('This functionality has not been implemented yet.')
        # TODO: Check how AIImageBuilder works with the
        # AIImage @dataclass and imitate it
        return VideoFileClip('test.mp4')



"""
A term is some text that we want to be matched
in a text so, when detected, we do something
to improve the content of the video (adding
an audio or video, applying some effect, etc.)
so we have to define the term and what we want
to be done.

Here is an example that would show an sticker
image of 'lionel messi portrait' when the 
exact term 'Lionel Messi' is found in the
audio transcription.

"Lionel Messi": {
    "mode": "exact",
    "context": "generic",
    "enhancements": [
        {
            "type": "sticker",
            "keywords": "lionel messi portrait",
            "url": "",
            "filename": "",
            "mode": "overlay"
        }
    ]
}
"""
from yta_core.classes.enhancement_json import EnhancementJson
from yta_core.enums.start import ShortcodeStart
from yta_core.edition_manual.enums import EditionManualTermContext, EditionManualTermField, EditionManualTermMode
from yta_audio.voice.transcription.objects import AudioTranscription
from yta_general_utils.text.finder import TextFinder
from yta_general_utils.file.reader import FileReader
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.validator import PythonValidator
from typing import Union
from copy import copy

import json


class EditionManualTerm:
    """
    Class to describe the behaviour we want from
    this term when applying it over a video base.
    It can be transformed into an effect, a
    transition, a video and any other element 
    that will improve the video experience.

    More information:
    - https://www.notion.so/Diccionarios-de-mejora-155efcba8f0d44e0890b178effb3be84?pvs=4
    """

    term: str = None
    """
    The text of the term.
    """
    mode: Union[EditionManualTermMode, None] = None
    """
    The mode in which the term must be searched.
    """
    context: Union[EditionManualTermContext, str] = None
    """
    The context in which the term must be applied.

    This means, if this term is for a 'sport' context
    or similar, can only be used in that context.
    """
    _enhancements: list[EnhancementJson, dict] = None
    """
    The list of enhancements that must be applied when
    the term is found.
    """

    @property
    def as_dict(
        self
    ):
        """
        Get the instance as a dict in which there is
        only one unique key that is the 'term', and
        the values are its 'mode', 'context' and
        'enhancements' fields.
        """
        return {
            self.term: {
                EditionManualTermField.MODE.value: self.mode.value,
                EditionManualTermField.CONTEXT.value: self.context.value,
                # TODO: These enhancements have to be returned as dicts
                EditionManualTermField.ENHANCEMENTS.value: [
                    enhancement.as_dict
                    for enhancement in self.enhancements
                ]
            }
        }
    
    @property
    def as_json(
        self
    ):
        """
        Get the instance as a json.
        """
        return json.dumps(self.as_dict)

    @property
    def enhancements(
        self
    ):
        return self._enhancements
    
    @enhancements.setter
    def enhancements(
        self,
        enhancements: list[EnhancementJson, dict]
    ):
        # TODO: Make some checkings and improvements
        ParameterValidator.validate_mandatory_list_of_these_instances('enhancements', enhancements, [EnhancementJson, dict])

        # These enhancements should have been validated before when
        # the EditionManual is accepted, so now we consider them as
        # valids

        # Here 'enhancements' are only dicts

        # # We turn dicts to EnhancementElement if necessary
        # obj_enhancements = []
        # for enhancement in enhancements:
        #     if not isinstance(enhancement, EnhancementElement) and not issubclass(enhancement.__class__, EnhancementElement):
        #         obj_enhancements.append(EnhancementElement.get_class_from_type(enhancement['type'])(enhancement['type'], EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD, EnhancementElementDuration.SHORTCODE_CONTENT, enhancement['keywords'], enhancement.get('url', ''), enhancement.get('filename', ''), enhancement['mode']))
        #     else:
        #         obj_enhancements.append(enhancement)

        self._enhancements = enhancements

    def __init__(
        self,
        term: str,
        mode: Union[EditionManualTermMode, str],
        context: Union[EditionManualTermContext, str],
        enhancements: list[dict]
    ):
        mode = EditionManualTermMode.to_enum(mode)
        context = EditionManualTermContext.to_enum(context)

        self.term = term
        self.mode = mode
        self.context = context
        self.enhancements = enhancements

    @staticmethod
    def init_from_dict(
        term_key: str,
        term_content: dict
    ):
        """
        Initialize an instance from the given
        'term_key' and 'term_content'. The
        'term_content' must have the expected
        structure and the values must be valid
        for our system.

        An example would be:
        - term_key = 'Lionel Messi'
        - term_content = {
            'mode': 'inline',
            'context': 'any',
            'enhancements': []
        }
        """
        ParameterValidator.validate_mandatory_string('term_key', term_key, do_accept_empty = False)
        EditionManualTerm.validate(term_content)
        
        return EditionManualTerm(
            term = term_key,
            mode = term_content[EditionManualTermField.MODE.value],
            context = term_content[EditionManualTermField.CONTEXT.value],
            enhancements = term_content[EditionManualTermField.ENHANCEMENTS.value]
        )
    
    @staticmethod
    def validate(
        term_content: dict
    ):
        """
        Check that the provided 'term_content' has the
        mandatory attributes and that the values are
        valid as if it was an Enhancement written in the
        'guion' of a video.

        The 'term_content' must be the content associated
        to the term including not the term key.
        """
        all_mandatory_fields = EditionManualTermField.get_all_values()

        if (
            not PythonValidator.do_dict_has_keys(term_content, all_mandatory_fields) or
            len(term_content) != 3
        ):
            raise Exception(f'The "term_content" parameter provided must have only the {len(all_mandatory_fields)} fields we accept: {", ".join(all_mandatory_fields)}.')
        
        EditionManualTermMode.to_enum(term_content[EditionManualTermField.MODE.value])
        EditionManualTermContext.to_enum(term_content[EditionManualTermField.CONTEXT.value])

        enhancements = term_content[EditionManualTermField.ENHANCEMENTS.value]
        for enhancement in enhancements:
            # TODO: Validate enhancement structure as if 
            # it were in the 'guion' definition
            #EnhancementJson.validate(enhancement)
            pass

    # TODO: Maybe apply return type and a @dataclass (?)
    def search(
        self,
        transcription: AudioTranscription
    ):
        """
        Searches the term in the provided 'transcription' text
        and, if found, returns the corresponding Enhancements
        with the processed duration (if it is possible to
        process it).

        This method returns the list of enhancement elements
        that should be applied, according to the provided
        'transcription', as dict elements.
        """
        ParameterValidator.validate_mandatory_instance_of('transcription', transcription, AudioTranscription)

        terms_found = TextFinder.find_in_text(self.term, transcription.text, mode = self.mode)

        enhancements_found = []
        for term_found in terms_found:
            for term_enhancement in self.enhancements:
                # TODO: Handle 'enhancement' as @dataclass please
                # We will add some fields to the 'enhancement'
                enhancement_found = copy(term_enhancement)
                # TODO: If 'term' is just one word we cannot use this
                # strategy below
                if (
                    enhancement_found['start'] == ShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD.value and
                    term_found.start_index == term_found.end_index
                ):
                    enhancement_found['start'] = ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD.value

                # Here we need to transform string 'start' and 'duration'
                # into their real numeric values
                start = None
                duration = None

                if enhancement_found['start'] == ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = transcription[term_found.start_index]['start']
                    duration = transcription[term_found.end_index]['end'] - start
                elif enhancement_found['start'] == ShortcodeStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = (transcription[term_found.start_index]['start'] + transcription[term_found.start_index]['end']) / 2
                    duration = (transcription[term_found.end_index]['start'] + transcription[term_found.end_index]['end']) / 2 - start
                elif enhancement_found['start'] == ShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = transcription[term_found.start_index]['end']
                    duration = start - transcription[term_found.end_index]['start']

                if (
                    start is None or
                    duration is None
                ):
                    raise Exception('Something went wrong when applying a EditionManualTerm.')

                enhancement_found['start'] = start
                enhancement_found['duration'] = duration
                
                # TODO: Should I turn this into a Enhancement object (?)
                # It is a dict right here
                enhancements_found.append(enhancement_found)

        return enhancements_found



from yta_stock_downloader import StockVideoDownloader, StockImageDownloader
from yta_general_utils.programming.decorators import singleton
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union


__all__ = [
    'StockDownloader'
]

@singleton
class StockDownloader:
    """
    Singleton class.

    This object simplifies the access to stock videos
    from official platforms. It uses the different
    platforms APIs.
    """

    _do_ignore_ids: bool
    """
    A flag to indicate if previously downloaded ids 
    (that are stored in the internal attributes to
    ignore from each platform) have to be ignored or
    not.
    """

    @property
    def do_ignore_ids(
        self
    ) -> bool:
        return self._do_ignore_ids
    
    @do_ignore_ids.setter
    def do_ignore_ids(
        self,
        value: bool
    ) -> None:
        self._do_ignore_ids = value
        self._video_downloader._do_ignore_ids = value
        self._image_downloader._do_ignore_ids = value

    def __init__(
        self,
        do_ignore_ids: bool = True
    ):
        self._video_downloader = StockVideoDownloader(do_ignore_ids = do_ignore_ids)
        self._image_downloader = StockImageDownloader(do_ignore_ids = do_ignore_ids)
        self._do_ignore_ids = do_ignore_ids

    def activate_ignore_repeated(
        self
    ) -> None:
        """
        Set as True the internal flag that indicates 
        that the ids of the videos that are downloaded
        after being activated have to be ignored in 
        the next downloads, so each video is downloaded
        only once.
        """
        self.do_ignore_ids = True

    def deactivate_ignore_repeated(
        self
    ) -> None:
        """
        Set as False the internal flag that indicates 
        that the ids of the videos that are downloaded
        after being activated have to be ignored in 
        the next downloads, so each video can be 
        downloaded an unlimited amount of times.
        """
        self.do_ignore_ids = True

    def reset(
        self
    ):
        """
        This method will empty the array that handles 
        the duplicated ids (if activated) to enable 
        downloading any resource found again. This
        method will also reset its platform-specific 
        downloader instances.
        """
        self._video_downloader.reset()
        self._image_downloader.reset()

    def download_video(
        self,
        keywords: str,
        do_randomize: bool = False,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download a video with the given 'keywords' from the
        stock platforms. The video will be the first found
        or a random one if 'do_randomize' is True.

        This method will return the filename of the video
        once it's been downloaded and stored locally, or
        None if no one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)
        
        download_method = (
            lambda query, output_filename: self._video_downloader.download_random(
                query = query,
                output_filename = output_filename
            )
            if do_randomize else
            lambda query, output_filename: self._video_downloader.download(
                query = query,
                output_filename = output_filename
            )
        )

        # TODO: What about the sound (?)
        try:
            video = download_method(
                query = keywords,
                output_filename = Output.get_filename(output_filename, FileTypeX.VIDEO)
            )
        except:
            pass

        # TODO: Should we try the 'download_image' if no video
        # available? Or should that strategy be built in the
        # part that requests it (?)

        # TODO: Should I return the FileReturn instead (?)
        return (
            video.filename
            if video is not None else
            None
        )
    
    def download_image(
        self,
        keywords: str,
        do_randomize: bool = False,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download an image with the given 'keywords' from the
        stock platforms. The image will be the first found
        or a random one if 'do_randomize' is True.

        This method will return the filename of the image
        once it's been downloaded and stored locally, or
        None if no one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)

        download_method = (
            lambda query, output_filename: self._image_downloader.download_random(
                query = query,
                output_filename = output_filename
            )
            if do_randomize else
            lambda query, output_filename: self._image_downloader.download(
                query = query,
                output_filename = output_filename
            )
        )

        try:
            image = download_method(
                query = keywords,
                output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
            )
        except:
            pass

        # TODO: Should I return the FileReturn instead (?)
        return (
            image.filename
            if image is not None else
            None
        )


from yta_core.builder import Builder
from yta_core.enums.field import EnhancementField
from yta_multimedia.greenscreen.custom.greenscreen import Greenscreen
from yta_general_utils.programming.validator.parameter import ParameterValidator


__all__ = [
    'GreenscreenBuilder'
]

class GreenscreenBuilder(Builder):
    """
    The builder of the GREENSCREEN type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        """
        Get an ImageGreenscreen or VideoGreenscreen instance
        initialized with the provided 'filename' or 'url'
        (filename has priority) or raises an Exception if
        something goes wrong.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        filename = enhancement.get(EnhancementField.FILENAME.value, None)
        url = enhancement.get(EnhancementField.URL.value, None)

        if url is None and filename is None:
            # TODO: I think this cannot happen
            raise Exception('No "url" nor "filename" provided and at least one is needed.')

        return (
            Greenscreen.init(url)
            if url and not filename else
            Greenscreen.init(filename)
        )


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Status(Enum):
    """
    Enum class to represent the different
    statuses an item can have.

    For internal use only. Use a more specific
    Enum class to interact with the different
    statuses.
    """

    TO_START = 'to_start'
    """
    It has not started yet.
    """
    IN_PROGRESS = 'in_progress'
    """
    It has not been built completely yet but it
    is being built.
    """
    FINISHED = 'finished'
    """
    It has been built and applied completely.
    """

class ProjectStatus(Enum):
    """
    Enum class to represent the different
    statuses that a project can have.
    """

    TO_START = _Status.TO_START.value
    """
    It has not started yet.
    """
    IN_PROGRESS = _Status.IN_PROGRESS.value
    """
    It has not been built completely yet but it
    is being built.
    """
    FINISHED = _Status.FINISHED.value
    """
    It has been built and applied completely.
    """

class SegmentStatus(Enum):
    """
    The current segment status defined by this string.
    """

    TO_START = _Status.TO_START.value
    """
    It has not started yet.
    """
    IN_PROGRESS = _Status.IN_PROGRESS.value
    """
    It has not been built completely yet but it
    is being built.
    """
    FINISHED = _Status.FINISHED.value
    """
    It has been built and applied completely.
    """

class EnhancementStatus(Enum):
    """
    The current enhancement status defined by this string.
    """

    TO_START = _Status.TO_START.value
    """
    It has not started yet.
    """
    IN_PROGRESS = _Status.IN_PROGRESS.value
    """
    It has not been built completely yet but it
    is being built.
    """
    FINISHED = _Status.FINISHED.value
    """
    It has been built and applied completely.
    """